# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Attribute MP_REACH_NLRI
"""

import struct

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep


class TunnelEncaps(Attribute):
    """
        Tunnel_Encapsulation_Attribute coding information
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |    Tunnel Type (2 Octets)     |        Length (2 Octets)      |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                                                               |
        |                             Value                             |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        the field of Value consists of sub-TLVs

        sub-TLVs coding information
        +-----------------------------------+
        |      Sub-TLV Type (1 Octet)       |
        +-----------------------------------+
        |     Sub-TLV Length (1 Octet)      |
        +-----------------------------------+
        |     Sub-TLV Value (Variable)      |
        |                                   |
        +-----------------------------------+
    """

    ID = AttributeID.Tunnel_Encapsulation_Attribute
    FLAG = AttributeFlag.OPTIONAL + AttributeFlag.TRANSITIVE + AttributeFlag.EXTENDED_LENGTH

    @classmethod
    def construct_optional_label_sid(cls, value):
        items = value.keys()
        label = value['label'] << 12
        tra_cls = 0
        btt_stack = 0
        ttl = 0
        if 'TC' in items:
            tra_cls = value['TC'] << 9
        else:
            tra_cls = 0
        if 'S' in items:
            btt_stack = value['S'] << 8
        else:
            btt_stack = 0
        if 'TTL' in items:
            ttl = value['TTL']
        else:
            ttl = 255
        return label + tra_cls + btt_stack + ttl

    @classmethod
    def construct_weight_and_seg(cls, segment_list):
        weight_hex = b''
        if bgp_cons.BGPSUB_TLV_WEIGHTED in segment_list.keys():
            weight_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_WEIGHTED) + struct.pack('!B', 6) + b'\x00\x00' +\
                struct.pack('!I', segment_list[bgp_cons.BGPSUB_TLV_WEIGHTED])
        seg_hex = b''
        for seg in segment_list[bgp_cons.BGPSUB_TLV_SID]:
            seg_type = int(seg.keys()[0])
            if seg_type == bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_MPLS:
                value = seg[seg.keys()[0]]
                sum_value = cls.construct_optional_label_sid(value)
                seg_hex += struct.pack('!B', bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_MPLS) + struct.pack('!B', 6) + b'\x00\x00' +\
                    struct.pack('!I', sum_value)
            # 3
            elif seg_type == bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_SID:
                value = seg[seg.keys()[0]]
                ipv4_node = value['node']
                if "SID" not in value.keys():
                    seg_hex += struct.pack('!B', bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_SID) + struct.pack('!B', 6) + b'\x00\x00' +\
                        netaddr.IPAddress(ipv4_node).packed
                else:
                    opt_sid = value['SID']
                    sum_value = cls.construct_optional_label_sid(opt_sid)
                    seg_hex += struct.pack('!B', bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_SID) + struct.pack('!B', 10) + b'\x00\x00' +\
                        netaddr.IPAddress(ipv4_node).packed + struct.pack('!I', sum_value)
            # 5
            elif seg_type == bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_INDEX_SID:
                value = seg[seg.keys()[0]]
                local_int = value['interface']
                ipv4_node = value['node']
                if "SID" not in value.keys():
                    seg_hex += struct.pack('!B', bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_INDEX_SID) +\
                        struct.pack('!B', 10) + b'\x00\x00' + struct.pack('!I', local_int) +\
                        netaddr.IPAddress(ipv4_node).packed
                else:
                    opt_sid = value['SID']
                    sum_value = cls.construct_optional_label_sid(opt_sid)
                    seg_hex += struct.pack('!B', bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_INDEX_SID) +\
                        struct.pack('!B', 14) + b'\x00\x00' + struct.pack('!I', local_int) +\
                        netaddr.IPAddress(ipv4_node).packed + struct.pack('!I', sum_value)
            # 6
            elif seg_type == bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_ADDR_SID:
                value = seg[seg.keys()[0]]
                local_ipv4 = value['local']
                remote_ipv4 = value['remote']
                if "SID" not in value.keys():
                    seg_hex += struct.pack('!B', bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_ADDR_SID) + struct.pack('!B', 10) +\
                        b'\x00\x00' + netaddr.IPAddress(local_ipv4).packed + netaddr.IPAddress(remote_ipv4).packed
                else:
                    opt_sid = value['SID']
                    sum_value = cls.construct_optional_label_sid(opt_sid)
                    seg_hex += struct.pack('!B', bgp_cons.BGP_SRTE_SEGMENT_SUBTLV_IPV4_ADDR_SID) + struct.pack('!B', 14) +\
                        b'\x00\x00' + netaddr.IPAddress(local_ipv4).packed + netaddr.IPAddress(remote_ipv4).packed +\
                        struct.pack('!I', sum_value)
        return weight_hex, seg_hex

    @classmethod
    def construct(cls, value):

        """Construct a attribute

        :param value: python dictionary
        {
            "0": "old",
            "6": 100,
            "7": 25102,
            "128": [
                {
                    "9": 10,
                    "1": [
                        {
                            "1": {
                                "label": 2000
                            }
                        },
                        {
                            "3": {
                                "node": "10.1.1.1",
                                "SID": {
                                    "label": 3000,
                                    "TC": 0,
                                    "S": 0,
                                    "TTL": 255
                                }
                            }
                        }
                    ]
                }
            ]
        }
        """
        policy_hex = b''
        policy = value
        data = dict([(int(l), r) for (l, r) in policy.items()])
        policy_value_hex = b''
        items = data.keys()
        if bgp_cons.BGP_BSID_PREFERENCE_OLD_OR_NEW not in items:
            raise excep.ConstructAttributeFailed(
                reason='failed to construct attributes: %s' % 'please provide the value of TLV encoding',
                data={}
            )
        for type_tmp in items:
            if type_tmp == bgp_cons.BGP_BSID_PREFERENCE_OLD_OR_NEW:
                # format of binding sid is same with the mpls label
                # the high order 20 bit was truly used as the bsid
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                # |          Label                        |           ...         |
                # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                # old ios
                if data[type_tmp] == 'old':
                    # Preference Sub-TLV
                    if bgp_cons.BGPSUB_TLV_PREFERENCE not in items:
                        if bgp_cons.BGPSUB_TLV_PREFERENCE_NEW in items:
                            policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_PREFERENCE) + struct.pack('!B', 6) + \
                                b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_PREFERENCE_NEW])
                    else:
                        policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_PREFERENCE) + struct.pack('!B', 6) + \
                            b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_PREFERENCE])
                    # Binding SID Sub-TLV
                    if bgp_cons.BGPSUB_TLV_BINDGINGSID not in items:
                        if bgp_cons.BGPSUB_TLV_BINDGINGSID_NEW not in items:
                            policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_BINDGINGSID) + struct.pack('!B', 2) +\
                                b'\x00\x00'
                        else:
                            policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_BINDGINGSID) + struct.pack('!B', 6) +\
                                b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_BINDGINGSID_NEW] << 12)
                    else:
                        policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_BINDGINGSID) + struct.pack('!B', 6) +\
                            b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_BINDGINGSID] << 12)
                # new ios
                elif data[type_tmp] == 'new':
                    # Preference Sub-TLV
                    if bgp_cons.BGPSUB_TLV_PREFERENCE not in items:
                        if bgp_cons.BGPSUB_TLV_PREFERENCE_NEW in items:
                            policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_PREFERENCE_NEW) + struct.pack('!B', 6) + \
                                b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_PREFERENCE_NEW])
                    else:
                        policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_PREFERENCE_NEW) + struct.pack('!B', 6) + \
                            b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_PREFERENCE])
                    # Binding SID Sub-TLV
                    if bgp_cons.BGPSUB_TLV_BINDGINGSID not in items:
                        if bgp_cons.BGPSUB_TLV_BINDGINGSID_NEW not in items:
                            policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_BINDGINGSID_NEW) + struct.pack('!B', 2) +\
                                b'\x00\x00'
                        else:
                            policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_BINDGINGSID_NEW) + struct.pack('!B', 6) +\
                                b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_BINDGINGSID_NEW] << 12)
                    else:
                        policy_value_hex += struct.pack('!B', bgp_cons.BGPSUB_TLV_BINDGINGSID_NEW) + struct.pack('!B', 6) +\
                            b'\x00\x00' + struct.pack('!I', data[bgp_cons.BGPSUB_TLV_BINDGINGSID] << 12)
                else:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes: %s' % 'TLV encoding must be one value of new or old',
                        data={}
                    )
            if type_tmp == bgp_cons.BGPSUB_TLV_SIDLIST:
                # Sub_TLV segment list
                seg_list_hex = b''
                for seg_list in data[type_tmp]:
                    segment_list = dict([(int(l), r) for (l, r) in seg_list.items()])
                    weight_hex, seg_hex = cls.construct_weight_and_seg(segment_list)
                    seg_list_hex += struct.pack('!B', type_tmp) + struct.pack('!H', len(weight_hex) + len(seg_hex) + 1) +\
                        b'\x00' + weight_hex + seg_hex
                policy_value_hex += seg_list_hex
        policy_hex += struct.pack('!H', bgp_cons.BGP_TUNNEL_ENCAPS_SR_TE_POLICY_TYPE) +\
            struct.pack('!H', len(policy_value_hex)) + policy_value_hex
        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) + struct.pack('!H', len(policy_hex)) + policy_hex



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Attribute MP_UNREACH_NLRI
"""

import struct

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute.nlri.ipv4_mpls_vpn import IPv4MPLSVPN
from yabgp.message.attribute.nlri.ipv6_mpls_vpn import IPv6MPLSVPN
from yabgp.message.attribute.nlri.ipv4_flowspec import IPv4FlowSpec
from yabgp.message.attribute.nlri.ipv6_unicast import IPv6Unicast
from yabgp.message.attribute.nlri.evpn import EVPN
from yabgp.message.attribute.nlri.linkstate import BGPLS
from yabgp.message.attribute.nlri.ipv4_srte import IPv4SRTE
from yabgp.common import afn
from yabgp.common import safn
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons


class MpUnReachNLRI(Attribute):

    """
        This is an optional non-transitive attribute that can be used for the
    purpose of withdrawing multiple unfeasible routes from service.
        An UPDATE message that contains the MP_UNREACH_NLRI is not required
    to carry any other path attributes.

        MP_UNREACH_NLRI coding information
        +---------------------------------------------------------+
        | Address Family Identifier (2 octets)                    |
        +---------------------------------------------------------+
        | Subsequent Address Family Identifier (1 octet)          |
        +---------------------------------------------------------+
        | Withdrawn Routes (variable)                             |
        +---------------------------------------------------------+
    """

    ID = AttributeID.MP_UNREACH_NLRI
    FLAG = AttributeFlag.OPTIONAL + AttributeFlag.EXTENDED_LENGTH

    @classmethod
    def parse(cls, value):
        try:
            afi, safi = struct.unpack('!HB', value[0:3])
        except Exception:
            raise excep.UpdateMessageError(sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                                           data='')

        nlri_bin = value[3:]

        # for IPv4
        if afi == afn.AFNUM_INET:

            # VPNv4
            if safi == safn.SAFNUM_LAB_VPNUNICAST:
                nlri = IPv4MPLSVPN.parse(nlri_bin, iswithdraw=True)
                return dict(afi_safi=(afi, safi), withdraw=nlri)
            # BGP flow spec
            elif safi == safn.SAFNUM_FSPEC_RULE:
                # if nlri length is greater than 240 bytes, it is encoded over 2 bytes
                withdraw_list = []
                while nlri_bin:
                    length = ord(nlri_bin[0])
                    if length >> 4 == 0xf and len(nlri_bin) > 2:
                        length = struct.unpack('!H', nlri_bin[:2])[0]
                        nlri_tmp = nlri_bin[2: length + 2]
                        nlri_bin = nlri_bin[length + 2:]
                    else:
                        nlri_tmp = nlri_bin[1: length + 1]
                        nlri_bin = nlri_bin[length + 1:]
                    nlri = IPv4FlowSpec.parse(nlri_tmp)
                    if nlri:
                        withdraw_list.append(nlri)

                return dict(afi_safi=(afi, safi), withdraw=withdraw_list)
            else:
                return dict(afi_safi=(afn.AFNUM_INET, safi), withdraw=repr(nlri_bin))
        # for ipv6
        elif afi == afn.AFNUM_INET6:
            # for ipv6 unicast
            if safi == safn.SAFNUM_UNICAST:
                return dict(afi_safi=(afi, safi), withdraw=IPv6Unicast.parse(nlri_data=nlri_bin))
            elif safi == safn.SAFNUM_LAB_VPNUNICAST:
                return dict(afi_safi=(afi, safi), withdraw=IPv6MPLSVPN.parse(value=nlri_bin, iswithdraw=True))
            else:
                return dict(afi_safi=(afi, safi), withdraw=repr(nlri_bin))
        # for l2vpn
        elif afi == afn.AFNUM_L2VPN:
            # for evpn
            if safi == safn.SAFNUM_EVPN:
                return dict(afi_safi=(afi, safi), withdraw=EVPN.parse(nlri_data=nlri_bin))
            else:
                return dict(afi_safi=(afi, safi), withdraw=repr(nlri_bin))
        # BGP LS
        elif afi == afn.AFNUM_BGPLS:
            if safi == safn.SAFNUM_BGPLS:
                withdraw = BGPLS.parse(nlri_bin)
                return dict(afi_safi=(afi, safi), withdraw=withdraw)
            else:
                pass
        else:
            return dict(afi_safi=(afi, safi), withdraw=repr(nlri_bin))

    @classmethod
    def construct(cls, value):

        """Construct a attribute

        :param value: python dictionary
        {'afi_safi': (1,128),
         'withdraw': []
        """
        afi, safi = value['afi_safi']
        if afi == afn.AFNUM_INET:
            if safi == safn.SAFNUM_LAB_VPNUNICAST:  # MPLS VPN
                nlri = IPv4MPLSVPN.construct(value['withdraw'], iswithdraw=True)
                if nlri:
                    attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + nlri
                    return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                        + struct.pack('!H', len(attr_value)) + attr_value
                else:
                    return None
            elif safi == safn.SAFNUM_FSPEC_RULE:
                try:
                    nlri_list = value.get('withdraw') or []
                    if not nlri_list:
                        return None
                    nlri_hex = b''
                    nlri_hex += IPv4FlowSpec.construct(value=nlri_list)
                    attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + nlri_hex
                    return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                        + struct.pack('!H', len(attr_value)) + attr_value

                except Exception:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes',
                        data=value
                    )
            elif safi == safn.SAFNUM_SRTE:
                try:
                    nlri_list = value.get('withdraw') or {}
                    if not nlri_list:
                        return None
                    nlri_hex = b''
                    nlri_hex += IPv4SRTE.construct(data=value['withdraw'])
                    attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + nlri_hex
                    return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                        + struct.pack('!H', len(attr_value)) + attr_value
                except Exception:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes',
                        data=value
                    )
            else:
                raise excep.ConstructAttributeFailed(
                    reason='unsupport this sub address family',
                    data=value)
        elif afi == afn.AFNUM_INET6:
            if safi == safn.SAFNUM_UNICAST:
                nlri = IPv6Unicast.construct(nlri_list=value['withdraw'])
                if nlri:
                    attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + nlri
                    return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                        + struct.pack('!H', len(attr_value)) + attr_value
            elif safi == safn.SAFNUM_LAB_VPNUNICAST:
                nlri = IPv6MPLSVPN.construct(value=value['withdraw'], iswithdraw=True)
                if nlri:
                    attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + nlri
                    return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                        + struct.pack('!H', len(attr_value)) + attr_value
                else:
                    return None
        # for l2vpn
        elif afi == afn.AFNUM_L2VPN:
            # for evpn
            if safi == safn.SAFNUM_EVPN:
                nlri = EVPN.construct(nlri_list=value['withdraw'])
                if nlri:
                    attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + nlri
                    return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                        + struct.pack('!H', len(attr_value)) + attr_value
            else:
                return None
        else:
            raise excep.ConstructAttributeFailed(
                reason='unsupport this sub address family',
                data=value)



# Copyright 2015-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.message.attribute.nlri.labeled_unicast.ipv6 import IPv6LabeledUnicast


class TestIPv6LabeledUnicast(unittest.TestCase):

    def test_construct(self):
        nlri_list = [
            {'label': [91], 'prefix': '2001:2121::1/128'},
            {'label': [92], 'prefix': '::2001:2121:1:0/64'},
            {'label': [93], 'prefix': '2001:4837:1821::2/127'}]
        nlri_hex = b'\x98\x00\x05\xb1\x20\x01\x21\x21\x00\x00\x00\x00\x00\x00\x00' \
            '\x00\x00\x00\x00\x01\x58\x00\x05\xc1\x20\x01\x21\x21\x00\x01' \
            '\x00\x00\x97\x00\x05\xd1\x20\x01\x48\x37\x18\x21\x00\x00\x00' \
            '\x00\x00\x00\x00\x00\x00\x02'

        self.assertEqual(nlri_hex, IPv6LabeledUnicast.construct(nlri_list))

    def test_parse(self):
        nlri_list = [
            {'label': [91], 'prefix': '2001:2121::1/128'},
            {'label': [92], 'prefix': '::2001:2121:1:0/64'},
            {'label': [93], 'prefix': '2001:4837:1821::2/127'}]
        nlri_hex = b'\x98\x00\x05\xb1\x20\x01\x21\x21\x00\x00\x00\x00\x00\x00\x00' \
            '\x00\x00\x00\x00\x01\x58\x00\x05\xc1\x20\x01\x21\x21\x00\x01' \
            '\x00\x00\x97\x00\x05\xd1\x20\x01\x48\x37\x18\x21\x00\x00\x00' \
            '\x00\x00\x00\x00\x00\x00\x02'

        self.assertEqual(nlri_list, IPv6LabeledUnicast.parse(nlri_hex))


if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http:\\www.apache.org\licenses\LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.message.attribute.nlri.linkstate import BGPLS


class TestBGPLSEPE(unittest.TestCase):

    def test_parse(self):
        self.maxDiff = None
        data_bin = b"\x00\x02\x00\x51\x07\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00" \
                   b"\x18\x02\x00\x00\x04\x00\x00\x00\xc8\x02\x01\x00\x04\x00\x00\x00" \
                   b"\x00\x02\x05\x00\x04\x00\x00\x00" \
                   b"\xc8\x01\x01\x00\x18\x02\x00\x00\x04\x00\x00\x01\x2c\x02\x01\x00" \
                   b"\x04\x00\x00\x00\x00\x02\x05\x00" \
                   b"\x04\x00\x00\x01\x2c\x01\x03\x00\x04\xc0\xa8\x04\x03\x01\x04\x00" \
                   b"\x04\xc0\xa8\x04\x04"
        data_dict = [
            {
                'type': 'link',
                'protocol_id': 7,
                'instances_id': 0,
                'descriptors': [
                    {
                        'type': 'local_node',
                        'value': {
                            'as_num': 200,
                            'bgpls_id': '0.0.0.0',
                            # 'bgp_router_id': '3.3.3.3',
                            'member_as_num': 200}},
                    {
                        'type': 'remote_node',
                        'value': {
                            'as_num': 300,
                            'bgpls_id': '0.0.0.0',
                            # 'bgp_router_id': '4.4.4.4',
                            'member_as_num': 300}},
                    {
                        'type': 'link_local_ipv4',
                        'value': '192.168.4.3'},
                    {
                        'type': 'link_remote_ipv4',
                        'value': '192.168.4.4'},
                ]
            }
        ]
        self.assertEqual(data_dict, BGPLS.parse(data_bin))



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.message.attribute.nlri.ipv4_srte import IPv4SRTE


class TestIPv4SRTE(unittest.TestCase):

    def test_construct(self):
        nlri_bin = b'\x60\x00\x00\x00\x00\x00\x00\x00\x0a\xc0\xa8\x05\x07'
        nlri = {"distinguisher": 0, "color": 10, "endpoint": "192.168.5.7"}
        self.assertEqual(nlri_bin, IPv4SRTE.construct(nlri))

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" basic config """

import logging
import sys

from oslo_config import cfg

CONF = cfg.CONF


BGP_CONFIG_OPTS = [
    cfg.IntOpt('peer_start_interval',
               default=10,
               help='The interval to start each BGP peer'),
    cfg.BoolOpt('four_bytes_as',
                default=True,
                help='If support 4bytes AS'),
    cfg.BoolOpt('route_refresh',
                default=True,
                help='If support sending and receiving route refresh message'),
    cfg.BoolOpt('cisco_route_refresh',
                default=True,
                help='If support sending and receiving cisco route refresh message'),
    cfg.BoolOpt('enhanced_route_refresh',
                default=True,
                help='If support enhanced route refresh'),
    cfg.StrOpt('add_path',
               choices=['ipv4_send', 'ipv4_receive', 'ipv4_both'],
               help='BGP additional path feature and supported address family'),
    cfg.BoolOpt('graceful_restart',
                default=True,
                help='if support graceful restart'),
    cfg.BoolOpt('cisco_multi_session',
                default=True,
                help='if support cisco multi session'),
    cfg.DictOpt('running_config',
                default={},
                help='The running configuration for BGP')
]

CONF.register_opts(BGP_CONFIG_OPTS, group='bgp')

BGP_PEER_CONFIG_OPTS = [
    cfg.IntOpt('remote_as',
               help='The remote BGP peer AS number'),
    cfg.IntOpt('local_as',
               help='The Local BGP AS number'),
    cfg.IPOpt('remote_addr',
              help='The remote address of the peer'),
    cfg.IPOpt('local_addr',
              default='0.0.0.0',
              help='The local address of the BGP'),
    cfg.StrOpt('md5',
               help='The MD5 string use to auth',
               secret=True),
    cfg.ListOpt('afi_safi',
                default=['ipv4'],
                help='The Global config for address family and sub address family')
]

CONF.register_cli_opts(BGP_PEER_CONFIG_OPTS, group='bgp')

LOG = logging.getLogger(__name__)


def get_bgp_config():
    """
    Get BGP running config
    :return:
    """
    # check bgp configuration from CLI input
    LOG.info('Try to load BGP configuration from CLI input')
    if CONF.bgp.local_as and CONF.bgp.remote_as and CONF.bgp.local_addr and CONF.bgp.remote_addr:
        CONF.bgp.running_config = {
            'remote_as': CONF.bgp.remote_as,
            'remote_addr': CONF.bgp.remote_addr,
            'local_as': CONF.bgp.local_as,
            'local_addr': CONF.bgp.local_addr,
            'md5': CONF.bgp.md5,
            'afi_safi': CONF.bgp.afi_safi,
            'capability': {
                'local': {
                    'four_bytes_as': CONF.bgp.four_bytes_as,
                    'route_refresh': CONF.bgp.route_refresh,
                    'cisco_route_refresh': CONF.bgp.cisco_route_refresh,
                    'enhanced_route_refresh': CONF.bgp.enhanced_route_refresh,
                    'graceful_restart': CONF.bgp.graceful_restart,
                    'cisco_multi_session': CONF.bgp.cisco_multi_session,
                    'add_path': CONF.bgp.add_path},
                'remote': {}
            }
        }
    else:
        LOG.error('Please provide enough parameters!')
        sys.exit()



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri.evpn import EVPN


class TestEVPN(unittest.TestCase):
    def test_parse_mac_ip_adv(self):
        data_hex = b'\x02\x25\x00\x01\xac\x11\x00\x03\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x6c\x30\x00\x11\x22\x33\x44\x55\x20\x0b\x0b\x0b\x01\x00\x00\x00'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_MAC_IP_ADVERTISEMENT,
            'value': {
                'rd': '172.17.0.3:2',
                'mac': '00-11-22-33-44-55',
                'eth_tag_id': 108,
                'esi': 0,
                'ip': '11.11.11.1',
                'label': [0]}
        }]
        self.assertEqual(data_list, EVPN.parse(data_hex))

    def test_construct_mac_ip_adv(self):
        data_hex = b'\x02\x25\x00\x01\xac\x11\x00\x03\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x6c\x30\x00\x11\x22\x33\x44\x55\x20\x0b\x0b\x0b\x01\x00\x00\x00'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_MAC_IP_ADVERTISEMENT,
            'value': {
                'rd': '172.17.0.3:2',
                'mac': '00-11-22-33-44-55',
                'eth_tag_id': 108,
                'esi': 0,
                'ip': '11.11.11.1',
                'label': [0]}
        }]
        self.assertEqual(data_hex, EVPN.construct(data_list))

    def test_parse_eth_auto_dis(self):
        data_hex = b'\x01\x19\x00\x01\x01\x01\x01\x01\x80\x63\x00\x00\x00' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x00\xa1'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_ETHERNET_AUTO_DISCOVERY,
            'value': {
                'rd': '1.1.1.1:32867',
                'esi': 0,
                'eth_tag_id': 100,
                'label': [10]
            }
        }]
        self.assertEqual(data_list, EVPN.parse(data_hex))

    def test_construct_eth_auto_dis(self):
        data_hex = b'\x01\x19\x00\x01\x01\x01\x01\x01\x80\x63\x00\x00\x00' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x00\xa1'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_ETHERNET_AUTO_DISCOVERY,
            'value': {
                'rd': '1.1.1.1:32867',
                'esi': 0,
                'eth_tag_id': 100,
                'label': [10]
            }
        }]
        self.assertEqual(data_hex, EVPN.construct(data_list))

    def test_parse_in_mul_eth_tag(self):
        data_hex = b'\x03\x11\x00\x01\xac\x10\x00\x01\x17\x10\x00\x00\x00\x64\x20\xc0\xa8\x00\x01'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_INCLUSIVE_MULTICAST_ETHERNET_TAG,
            'value': {
                'rd': '172.16.0.1:5904',
                'eth_tag_id': 100,
                'ip': '192.168.0.1'
            }
        }]
        self.assertEqual(data_list, EVPN.parse(data_hex))

    def test_construct_in_mul_eth_tag(self):
        data_hex = b'\x03\x11\x00\x01\xac\x10\x00\x01\x17\x10\x00\x00\x00\x64\x20\xc0\xa8\x00\x01'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_INCLUSIVE_MULTICAST_ETHERNET_TAG,
            'value': {
                'rd': '172.16.0.1:5904',
                'eth_tag_id': 100,
                'ip': '192.168.0.1'
            }
        }]
        self.assertEqual(data_hex, EVPN.construct(data_list))

    def test_parse_eth_segment(self):
        data_hex = b'\x04\x17\x00\x01\xac\x10\x00\x01\x17\x10\x00\x00' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x20\xc0\xa8\x00\x01'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_ETHERNET_SEGMENT,
            'value': {
                'rd': '172.16.0.1:5904',
                'esi': 0,
                'ip': '192.168.0.1'
            }
        }]
        self.assertEqual(data_list, EVPN.parse(data_hex))

    def test_construct_eth_segment(self):
        data_hex = b'\x04\x17\x00\x01\xac\x10\x00\x01\x17\x10\x00\x00' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x20\xc0\xa8\x00\x01'
        data_list = [{
            'type': bgp_cons.BGPNLRI_EVPN_ETHERNET_SEGMENT,
            'value': {
                'rd': '172.16.0.1:5904',
                'esi': 0,
                'ip': '192.168.0.1'
            }
        }]
        self.assertEqual(data_hex, EVPN.construct(data_list))

    def test_parse_ip_route_prefix_v4(self):
        data_hex = b'\x05\x22\x00\x02\x00\x01\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x01\x18\x01\x01\x01\x00\x01\x01\x01\x01\x00\x00\xa1'
        data_list = [{
            'type': 5,
            'value': {
                'esi': 0,
                'eth_tag_id': 1,
                'gateway': '1.1.1.1',
                'label': [10],
                'prefix': '1.1.1.0/24',
                'rd': '65536:2'}}]
        self.assertEqual(data_list, EVPN.parse(data_hex))

    def test_construct_ip_route_prefix_v4(self):
        data_hex = b'\x05\x22\x00\x02\x00\x01\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x01\x18\x01\x01\x01\x00\x01\x01\x01\x01\x00\x00\xa1'
        data_list = [{
            'type': 5,
            'value': {
                'esi': 0,
                'eth_tag_id': 1,
                'gateway': '1.1.1.1',
                'label': [10],
                'prefix': '1.1.1.0/24',
                'rd': '65536:2'}}]
        self.assertEqual(data_hex, EVPN.construct(data_list))

    def test_parse_ip_route_prefix_v6(self):
        data_hex = b'\x05' \
                   b'\x3a' \
                   b'\x00\x02\x00\x01\x00\x00\x00\x02' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x01' \
                   b'\x40' \
                   b'\x20\x01\x32\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01' \
                   b'\x20\x01\x32\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01' \
                   b'\x00\x00\xa1'
        data_list = [{
            'type': 5,
            'value': {
                'esi': 0,
                'eth_tag_id': 1,
                'gateway': '2001:3232::1',
                'label': [10],
                'prefix': '2001:3232::1/64',
                'rd': '65536:2'}}]
        self.assertEqual(data_list, EVPN.parse(data_hex))

    def test_construct_ip_route_prefix_v6(self):
        data_hex = b'\x05' \
                   b'\x3a' \
                   b'\x00\x02\x00\x01\x00\x00\x00\x02' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x01' \
                   b'\x40' \
                   b'\x20\x01\x32\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01' \
                   b'\x20\x01\x32\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01' \
                   b'\x00\x00\xa1'
        data_list = [{
            'type': 5,
            'value': {
                'esi': 0,
                'eth_tag_id': 1,
                'gateway': '2001:3232::1',
                'label': [10],
                'prefix': '2001:3232::1/64',
                'rd': '65536:2'}}]
        self.assertEqual(data_hex, EVPN.construct(data_list))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright 2015-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" BGP EVPN Route import
"""

from __future__ import print_function
import os
import time
import logging
import json
import traceback
import sys

from pymongo import MongoClient


CONFIG = {}

# constant
BGP_OPEN = 1
BGP_UPDATE = 2
BGP_NOTIFICATION = 3
BGP_KEEPALIVE = 4
BGP_ROUTE_REFRESH = 5

MONGO_COLLECTION_MAC_IP_VNI = 'MAC_IP_VNI'
MONGO_COLLECTION_MAC_IP_HIS = 'MAC_IP_HISTORY'
MONGO_COLLECTION_SNAPSHOT_SEQ = 'MSG_SEQ'

# mongodb
CONFIG['DB_NAME'] = os.environ.get('MONGODB_NAME', 'evpn')
CONFIG['DB_URL'] = os.environ.get('MONGODB_URL', 'mongodb://yabgp:yabgp@as-gerrit.cisco.com:27017,as-gerrit.cisco.com:27018,as-gerrit.cisco.com:27019')
CONFIG['MSG_PATH'] = os.environ.get('MSG_PATH', '~/data/bgp/11.11.11.11/msg')
# logging
handler = logging.FileHandler('evpn.log')
handler.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

# add the handlers to the logger
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)
LOG.addHandler(handler)


# database
db_client = MongoClient(host=CONFIG['DB_URL'])
dbh = db_client[CONFIG['DB_NAME']]


class MessageFileOperator(object):

    def __init__(self, msgfile_dir, lastseq=0, file_index=-1):
        """
        Get the file we wanted and located the line according to the sequence number.
        line of it.
        init message file object
        :param msgfile_dir: message file dir for this peer
        :param lastseq: last sequence number that read successfully, if `lastseq` = 0, then read
        from the first line of the first file, if `lastseq` = -1, then read from the first line from
        the last updated file. otherwise, try to find the exactly line of sequence number.
        :param file_index: locate file based its index
        """
        self.file_dir = msgfile_dir
        self.lastseq = lastseq
        self.file_index = file_index
        self.file_name = None
        self.last_line = ''
        self._f = None
        self.first_time_catchup_flag = False

    def locate_file_and_line(self):
        while True:
            self.file_name = self._locate_file()
            if self.file_name is not None:
                break
            # Wait until file exists
            time.sleep(20)
        self._f = self._locate_line(self.file_name, self.lastseq)

    def _locate_file(self):
        """
        locate the right message file
        """
        file_name = None
        LOG.info('Locate message file, when message seq = %s.' % self.lastseq)
        if not os.path.exists(self.file_dir):
            LOG.critical('The BGP data path does not exist, path=%s' % self.file_dir)
            sys.exit()
        self.file_list = os.listdir(self.file_dir)
        # delete dir
        dir_name_list = []
        for file_ in self.file_list:
            if os.path.isdir(os.path.join(self.file_dir, file_)):
                dir_name_list.append(file_)
        for file_ in dir_name_list:
            self.file_list.remove(file_)
        if not self.file_list:
            LOG.warning('There are no files! Will check again in 20 seconds.....')
            return None
        self.file_list.sort()

        if self.lastseq == 0:
            # return the first file
            file_name = os.path.join(self.file_dir, self.file_list[0])
            LOG.info('Locate file successfully, file = %s' % file_name)
            return file_name

        elif self.lastseq == -1:
            # return the last file
            file_name = os.path.join(self.file_dir, self.file_list[-1])
            LOG.info('Locate file successfully, file = %s' % file_name)
            return file_name

        # return the file contain the sequence number
        find_flag = False
        for file_ in self.file_list:
            # get the first line and last line, if sequence number is between the first line
            # and last line, then find_flag is True and return the file name
            file_name = os.path.join(self.file_dir, file_)
            first_line = {}
            last_line = {}
            with open(file_name, 'r') as f:
                t = 0
                while True:
                    first_line = next(f)
                    try:
                        first_line = json.loads(first_line)
                        break

                    except Exception as e:
                        LOG.error(e)
                        ex_str = traceback.format_exc()
                        LOG.error(ex_str)
                        time.sleep(120)
                        t += 1
                        if t > 10:
                            LOG.critical("Can not get first line of %s" % file_name)
                            sys.exit()
                offs = -100
                while True:
                    f.seek(offs, 2)
                    lines = f.readlines()
                    if len(lines) > 1:
                        try:
                            last_line = lines[-1]
                            last_line = json.loads(last_line)
                        except Exception, e:
                            LOG.exception(e.message)
                            last_line = lines[-2]
                            try:
                                last_line = json.loads(last_line)
                            except Exception as e:
                                LOG.error(e)
                                ex_str = traceback.format_exc()
                                LOG.error(ex_str)
                                LOG.critical("Can not get the last line of %s" % file_name)
                                sys.exit()
                        break
                    offs *= 2
            if first_line and last_line:
                if first_line['seq'] <= self.lastseq <= last_line['seq']:
                    find_flag = True
                    break
        if find_flag:
            LOG.info('Locate file successfully, file = %s' % file_name)
            return os.path.join(self.file_dir, file_name)
        else:
            LOG.critical('Can not locate message file, when seq=%s' % self.lastseq)
            sys.exit()

    @staticmethod
    def _locate_line(msgfile, lastseq=0):
        """Get bgp message file handles and seek to the position after seq number your input
        """

        # Open message file
        f = open(msgfile, "r")
        LOG.info('Open BGP message file: %s' % msgfile)
        if lastseq in [0, -1]:
            return f
        else:  # skip lastseq
            line = f.readline()
            while line:
                line_json = json.loads(line)
                if line_json['seq'] == lastseq:
                    break
                line = f.readline()
        return f

    @property
    def get_next_file(self):

        old_file_name = os.path.split(self.file_name)[-1]
        file_list = os.listdir(self.file_dir)
        # delete dirs
        dir_name_list = []
        for file_ in file_list:
            if os.path.isdir(os.path.join(self.file_dir, file_)):
                dir_name_list.append(file_)
        for file_ in dir_name_list:
            file_list.remove(file_)
        file_list.sort()
        if len(file_list) == 1:
            return None
        index = file_list.index(old_file_name)
        if index + 1 == len(file_list):
            return None
        try:
            new_file_name = file_list[index + 1]
        except Exception as e:
            LOG.error(e)
            ex_str = traceback.format_exc()
            LOG.error(ex_str)
            return None
        return os.path.join(self.file_dir, new_file_name)

    @property
    def readline(self):
        """
        read one line and return
        return None if no new line
        """

        line = self._f.readline()
        if line and line.endswith('\n'):
            self.last_line = line
            return line
        elif line:  # this line is still be writting
            last_len = len(line)
            self._f.seek(self._f.tell() - last_len)
            self.first_time_catchup_flag = True
            return None
        else:
            if not self.first_time_catchup_flag:
                self.first_time_catchup_flag = True
                LOG.info('first time catchup the file ending')
                return None
            time.sleep(1)
            # if need to open the next new file
            next_file = self.get_next_file
            if next_file:
                # try old file again
                line = self._f.readline()
                if line and line.endswith('\n'):
                    self.last_line = line
                    return line
                elif line:
                    last_len = len(line)
                    self._f.seek(self._f.tell() - last_len)
                    return None
                else:
                    # need to check the next file
                    with open(next_file, 'r') as f_next:
                        first_line = f_next.readline()
                        try:
                            first_seq = json.loads(first_line)['seq']
                            last_seq = json.loads(self.last_line)['seq']
                            LOG.debug('first seq of next file=%s, last seq of old file=%s' % (first_seq, last_seq))
                            if last_seq + 1 == first_seq:
                                # really need open next file
                                self._f.close()  # close old file
                                self._f = open(next_file, 'r')
                                LOG.info('Open next BGP message file: %s' % next_file)
                                self.file_name = next_file
                                return None
                            else:
                                f_next.close()
                                return None
                        except Exception as e:
                            LOG.error(e)
                            error_str = traceback.format_exc()
                            LOG.debug(error_str)
                            return None
            else:
                return None


class EVPNRoute(object):
    """
    evpn route process

    `MAC`, `IP`, `VNI`, `NH`, `VRF`, 'TEP'

    """
    def __init__():
        pass

    @staticmethod
    def clear():
        LOG.info('clear mongo collection %s', MONGO_COLLECTION_MAC_IP_VNI)
        dbh[MONGO_COLLECTION_MAC_IP_VNI].remove()

    @staticmethod
    def update(record):
        pass


    @staticmethod
    def get_l2_l3_vni(ext_comm_list):
        vni_list = [ext_comm for ext_comm in ext_comm_list if ext_comm[0] == 2]
        if len(vni_list) == 1:
            return None, vni_list[0][1].split(':')[-1]
        elif len(vni_list) >= 2:
            return vni_list[0][1].split(':')[-1], vni_list[1][1].split(':')[-1]
        else:
            return None, None


class EVPNHistory(object):
    pass


def main():
    LOG.info('-------> begin to import')
    file_handler = MessageFileOperator(CONFIG['MSG_PATH'] , lastseq=0)
    file_handler.locate_file_and_line()
    while True:
        line = file_handler.readline
        if not line:
            time.sleep(2)
            continue
        try:
            bgp_msg = json.loads(line)
            # message type
            if bgp_msg['type'] in [BGP_OPEN, BGP_NOTIFICATION]:
                # clear evpn mac ip table
                EVPNRoute.clear()
            elif bgp_msg['type'] == BGP_UPDATE:
                if bgp_msg['msg']['afi_safi'] != 'evpn':
                    continue

                # try to get l2 l3 vni
                l3vni, l2vni =  EVPNRoute.get_l2_l3_vni(bgp_msg['msg']['attr'].get('16') or [])
                tep = bgp_msg['msg']['attr'].get('9')
                # for nlri
                nlri_list = (bgp_msg['msg']['attr'].get('14') or {}).get('nlri') or []
                for prefix in nlri_list:
                    if prefix['type'] == 2:  # BGPNLRI_EVPN_MAC_IP_ADVERTISEMENT
                        ip = prefix['value'].get('ip')
                        mac = prefix['value'].get('mac')
                        if ip and mac:
                            print(ip, mac, l3vni, l2vni, tep)



        except Exception as e:
            LOG.error(traceback.format_exc())
            LOG.error('Message format error when using json.loads(line), line = %s detail: %s' % (line, e))
            continue

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Link State attribute"""

import unittest
from yabgp.message.attribute.linkstate.linkstate import LinkState


class TestLinkState(unittest.TestCase):

    def test_unpack(self):
        data_bin = b'\x04\x04\x00\x04\x02\x02\x02\x02\x04\x06\x00\x04\x01\x01\x01\x01' \
                   b'\x04\x40\x00\x04\x00\x00\x00\x00\x04\x41\x00\x04\x4c\xee\x6b\x28' \
                   b'\x04\x42\x00\x04\x00\x00\x00\x00\x04\x43\x00\x20\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x44\x00\x04' \
                   b'\x00\x00\x00\x0a\x04\x47\x00\x03\x00\x00\x0a\x04\x4b\x00\x07\x70' \
                   b'\x00\x00\x00\x00\x61\xa8\x04\x4b\x00\x07\x30\x00\x00\x00\x00\x61' \
                   b'\xa9'
        self.assertEqual(None, LinkState.unpack(data_bin).dict())



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function
import sys

from flask import Blueprint
import flask

from yabgp.agent import prepare_service
from yabgp.handler import BaseHandler


class APIHandler(object):

    blueprint = Blueprint('demo', __name__)
    url_prefix = '/v1'


@APIHandler.blueprint.route('/rib')
def root():
    """
    v1 api root. Get the api status.
    """
    intro = {
        'rib': []
    }
    return flask.jsonify(intro)


class CliHandler(BaseHandler):
    """demo handler implementation
    """

    def __init__(self):
        super(CliHandler, self).__init__()

    def init(self):
        pass

    def on_update_error(self, peer, timestamp, msg):
        print('[-] UPDATE ERROR,', msg)

    def route_refresh_received(self, peer, msg, msg_type):
        print('[+] ROUTE_REFRESH received,', msg)

    def keepalive_received(self, peer, timestamp):
        print('[+] KEEPALIVE received')

    def open_received(self, peer, timestamp, result):
        print('[+] OPEN received,', result)

    def update_received(self, peer, timestamp, msg):
        print('[+] UPDATE received,', msg)

    def notification_received(self, peer, msg):
        print('[-] NOTIFICATION received,', msg)

    def on_connection_lost(self, peer):
        print('[-] CONNECTION lost')

    def on_connection_failed(self, peer, msg):
        print('[-] CONNECTION failed,', msg)

    def on_established(self, peer, msg):
        print('[-] ESTABLISHED,', msg)


def main():
    try:
        cli_handler = CliHandler()
        prepare_service(handler=cli_handler, api_hander=APIHandler())
    except Exception as e:
        print(e)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
import binascii

import netaddr


class NLRI(object):

    @classmethod
    def parse(cls, *args):

        raise NotImplementedError

    @classmethod
    def construct(cls, *args):
        raise NotImplementedError

    @staticmethod
    def construct_prefix_v4(masklen, prefix_str):
        ip_hex = struct.pack('!I', netaddr.IPNetwork(prefix_str).value)
        if 16 < masklen <= 24:
            ip_hex = ip_hex[0:3]
        elif 8 < masklen <= 16:
            ip_hex = ip_hex[0:2]
        elif masklen <= 8:
            ip_hex = ip_hex[0:1]
        return ip_hex

    @staticmethod
    def construct_prefix_v6(prefix):
        mask = int(prefix.split('/')[1])
        prefix_hex = binascii.unhexlify(hex(netaddr.IPNetwork(prefix).ip)[2:])
        offset = mask / 8
        offset_re = mask % 8
        if offset == 0:
            return prefix_hex[0: 1]
        return prefix_hex[0: offset + offset_re]

    @classmethod
    def parse_mpls_label_stack(cls, data):
        labels = []
        while len(data) >= 3:
            label = struct.unpack('!L', b'\00'+data[:3])[0]
            data = data[3:]
            labels.append(label >> 4)
            if label & 0x001:
                break
        return labels

    @classmethod
    def construct_mpls_label_stack(cls, labels):
        data = b''
        last_label = labels[-1]
        for label in labels[:-1]:
            data += struct.pack('!L', label << 4)[1:]
        if last_label != 0:
            data += struct.pack('!L', (last_label << 4 | 1))[1:]
        else:
            data += b'\x00\x00\x00'
        return data



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/rfc7752#section-3.3.1.1
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |O|T|E|B|R|V| Rsvd|
#     +-+-+-+-+-+-+-+-+-+
#
#     +-----------------+-------------------------+------------+
#     |       Bit       | Description             | Reference  |
#     +-----------------+-------------------------+------------+
#     |       'O'       | Overload Bit            | [ISO10589] |
#     |       'T'       | Attached Bit            | [ISO10589] |
#     |       'E'       | External Bit            | [RFC2328]  |
#     |       'B'       | ABR Bit                 | [RFC2328]  |
#     |       'R'       | Router Bit              | [RFC5340]  |
#     |       'V'       | V6 Bit                  | [RFC5340]  |
#     | Reserved (Rsvd) | Reserved for future use |            |
#     +-----------------+-------------------------+------------+


@LinkState.register()
class NodeFlags(TLV):

    TYPE = 1024  # https://tools.ietf.org/html/rfc7752#section-3.3.1.1
    TYPE_STR = "node_flags"

    @classmethod
    def unpack(cls, value):
        """
        """
        valueByte = ord(value[0:1])
        O = valueByte >> 7
        T = (valueByte << 1) % 256 >> 7
        E = (valueByte << 2) % 256 >> 7
        B = (valueByte << 3) % 256 >> 7
        R = (valueByte << 4) % 256 >> 7
        V = (valueByte << 5) % 256 >> 7
        return cls(value={"O": O, "T": T, "E": E, "B": B, "R": R, "V": V})



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import binascii

from yabgp.tlv import TLV
from ..linkstate import LinkState

#     0                   1                   2                   3
#     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |               Type            |            Length             |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |     Flags     |   Algorithm   |           Reserved            |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |                       SID/Index/Label (variable)              |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


@LinkState.register()
class PrefixSID(TLV):
    """
    prefix sid
    """
    TYPE = 1158
    TYPE_STR = 'prefix_sid'

    @classmethod
    def unpack(cls, data):
        algorithm = ord(data[1:2])
        return cls(value={"algorithm": algorithm, "sid": int(binascii.b2a_hex(data[4:]), 16)})



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test IPv6 MPLS VPN NLRI """

import unittest

from yabgp.message.attribute.nlri.ipv6_mpls_vpn import IPv6MPLSVPN


class TestIPv6MPLSVPN(unittest.TestCase):

    def test_update_parse(self):
        nlri_hex = b'\x98\x00\x03\x61\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x00\x00\x12\x00\x04' \
                   b'\x98\x00\x03\x71\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x01\x00\x12\x00\x00'
        nlri_list = [
            {'label': [54], 'rd': '100:12', 'prefix': '2010:0:12:4::/64'},
            {'label': [55], 'rd': '100:12', 'prefix': '2010:1:12::/64'}
        ]
        self.assertEqual(nlri_list, IPv6MPLSVPN.parse(value=nlri_hex))

    def test_update_construct(self):
        nlri_hex = b'\x98\x00\x03\x61\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x00\x00\x12\x00\x04' \
                   b'\x98\x00\x03\x71\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x01\x00\x12\x00\x00'
        nlri_list = [
            {'label': [54], 'rd': '100:12', 'prefix': '2010:0:12:4::/64'},
            {'label': [55], 'rd': '100:12', 'prefix': '2010:1:12::/64'}
        ]
        self.assertEqual(nlri_hex, IPv6MPLSVPN.construct(value=nlri_list))

    def test_withdraw_parse(self):
        nlri_hex = b'\x98\x80\x00\x00\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x00\x00\x12\x00' \
                   b'\x04\x98\x80\x00\x00\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x01\x00\x12\x00\x00'
        nlri_list = [
            {'label': [524288], 'rd': '100:12', 'prefix': '2010:0:12:4::/64'},
            {'label': [524288], 'rd': '100:12', 'prefix': '2010:1:12::/64'}
        ]
        self.assertEqual(nlri_list, IPv6MPLSVPN.parse(value=nlri_hex, iswithdraw=True))

    def test_withdraw_construct(self):
        nlri_hex = b'\x98\x80\x00\x00\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x00\x00\x12\x00' \
                   b'\x04\x98\x80\x00\x00\x00\x00\x00\x64\x00\x00\x00\x0c\x20\x10\x00\x01\x00\x12\x00\x00'
        nlri_list = [
            {'rd': '100:12', 'prefix': '2010:0:12:4::/64'},
            {'rd': '100:12', 'prefix': '2010:1:12::/64'}
        ]
        self.assertEqual(nlri_hex, IPv6MPLSVPN.construct(value=nlri_list, iswithdraw=True))


if __name__ == '__main__':
    unittest.main()



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
import binascii
import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.common import constants as bgp_cons


class PMSITunnel(Attribute):
    """
    RFC 6514#section-5
    """
    ID = AttributeID.PMSI_TUNNEL
    FLAG = AttributeFlag.OPTIONAL + AttributeFlag.TRANSITIVE

    @classmethod
    def parse(cls, value, evpn_overlay=False):
        """
        +---------------------------------+
        |  Flags (1 octet)                |
        +---------------------------------+
        |  Tunnel Type (1 octets)         |
        +---------------------------------+
        |  MPLS Label (3 octets)          |
        +---------------------------------+
        |  Tunnel Identifier (variable)   |
        +---------------------------------+
        :param value raw hex value
        """
        flag = ord(value[0:1])
        tunnel_type = ord(value[1:2])
        if evpn_overlay:
            mpls_label = cls.parse_vni(value[2: 5])
        else:
            mpls_label = cls.parse_mpls_label(value[2: 5])
        tunnel_id = cls.parse_tunnel_id(tunel_type=tunnel_type, tunel_data=value[5:])
        return {
            'leaf_info_required': flag,
            'tunnel_type': tunnel_type,
            'mpls_label': [mpls_label],
            'tunnel_id': tunnel_id
        }

    @staticmethod
    def parse_mpls_label(data):
        label = struct.unpack('!L', b'\00'+data[:3])[0]
        return label >> 4

    @staticmethod
    def parse_vni(data):
        label = struct.unpack('!L', b'\00'+data[:3])[0]
        return label

    @staticmethod
    def parse_tunnel_id(tunel_type, tunel_data):
        if tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_NO_TUNNEL:
            # When the Tunnel Type is set to "No tunnel information present", the
            # PMSI Tunnel attribute carries no tunnel information (no Tunnel
            # Identifier).
            return None
        elif tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_RSVP_TE_P2MP:
            # When the Tunnel Type is set to RSVP - Traffic Engineering (RSVP-TE)
            # Point-to-Multipoint (P2MP) Label Switched Path (LSP), the Tunnel
            # Identifier is <Extended Tunnel ID, Reserved, Tunnel ID, P2MP ID> as
            # carried in the RSVP-TE P2MP LSP SESSION Object [RFC4875].
            pass
        elif tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_MLDP_P2MP:
            # When the Tunnel Type is set to multipoint Label Distribution Protocol
            # (mLDP) P2MP LSP, the Tunnel Identifier is a P2MP Forwarding
            # Equivalence Class (FEC) Element [mLDP].
            pass
        elif tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_PIM_SSM_TREE:
            # When the Tunnel Type is set to PIM-SSM tree, the Tunnel Identifier is
            # <P-Root Node Address, P-Multicast Group>.  The node that originates
            # the attribute MUST use the address carried in the P-Root Node Address
            # as the source IP address for the IP/GRE encapsulation of the MVPN
            # data.  The P-Multicast Group in the Tunnel Identifier of the Tunnel
            # attribute MUST NOT be expected to be the same group for all Intra-AS
            # A-D routes for the same MVPN.  According to [RFC4607], the group
            # address can be locally allocated by the originating PE without any
            # consideration for the group address used by other PE on the same
            # MVPN.
            pass
        elif tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_PIM_SM_TREE:
            # When the Tunnel Type is set to Protocol Independent Multicast -
            # Sparse Mode (PIM-SM) tree, the Tunnel Identifier is <Sender Address,
            # P-Multicast Group>.  The node that originated the attribute MUST use
            # the address carried in the Sender Address as the source IP address
            # for the IP/GRE (Generic Routing Encapsulation) encapsulation of the
            # MVPN data.
            pass
        elif tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_BIDIR_PIM_TREE:
            # When the Tunnel Type is set to BIDIR-PIM tree, the Tunnel Identifier
            # is <Sender Address, P-Multicast Group>.  The node that originated the
            # attribute MUST use the address carried in the Sender Address as the
            # source IP address for the IP/GRE encapsulation of the MVPN data.
            pass
        elif tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_INGRESS_REPL:
            # When the Tunnel Type is set to Ingress Replication, the Tunnel
            # Identifier carries the unicast tunnel endpoint IP address of the
            # local PE that is to be this PE's receiving endpoint address for the
            # tunnel.
            return str(netaddr.IPAddress(int(binascii.b2a_hex(tunel_data), 16)))
        elif tunel_type == bgp_cons.PMSI_TUNNEL_TYPE_MLDP_MP2MP:
            # When the Tunnel Type is set to mLDP Multipoint-to-Multipoint (MP2MP)
            # LSP, the Tunnel Identifier is an MP2MP FEC Element [mLDP].
            pass
        return 'not supported'

    @classmethod
    def construct(cls, value, evpn_overlay=False):
        """
        :param value:
            {
                "mpls_label": [1234],
                "tunnel_id": "192.168.10.10",
                "tunnel_type": 6,
                "leaf_info_required": 0
            }
        """
        pmsi_tunnel_hex = b''
        # Flags: (currently only Leaf Info Required is defined)
        pmsi_tunnel_hex += struct.pack('!B', int(value['leaf_info_required']))
        # Tunnel Type
        pmsi_tunnel_hex += struct.pack('!B', int(value['tunnel_type']))
        # MPLS Label
        pmsi_tunnel_hex += cls.construct_pmsi_label(evpn_overlay, value)
        # tunnel_type dictates the tunnel_id structure
        pmsi_tunnel_hex += cls.construct_tunnel_type(value)
        if pmsi_tunnel_hex:
            return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                + struct.pack('!B', len(pmsi_tunnel_hex)) + pmsi_tunnel_hex

    @staticmethod
    def construct_pmsi_label(evpn_overlay, value):
        if not evpn_overlay:
            # MPLS Label
            return struct.pack('!L', value['mpls_label'][0] << 4)[1:]
        else:
            # draft-ietf-bess-evpn-overlay-10#section-5.1.3 - label is used for 24 bit VNI
            if evpn_overlay['evpn'] & evpn_overlay['encap_ec']:
                if evpn_overlay['encap_value'] == bgp_cons.BGP_TUNNEL_ENCAPS_VXLAN:
                    return struct.pack('!L', value['mpls_label'][0])[1:]
                elif evpn_overlay['encap_value'] == bgp_cons.BGP_TUNNEL_ENCAPS_NVGRE:
                    return struct.pack('!L', value['mpls_label'][0])[1:]
            else:
                # MPLS Label
                return struct.pack('!L', value['mpls_label'][0] << 4)[1:]

    @staticmethod
    def construct_tunnel_type(value):
        if value['tunnel_type'] == bgp_cons.PMSI_TUNNEL_TYPE_INGRESS_REPL:
            return netaddr.IPAddress(value['tunnel_id']).packed
        else:
            return ''



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.message.attribute.nlri import NLRI


class TestNLRI(unittest.TestCase):

    def test_construct_prefix_vpnv6(self):
        prefix_hex = b'\x20\x10\x00\x00\x00\x12\x00\04'
        prefix_str = '2010:0:12:4::/64'
        self.assertEqual(prefix_hex, NLRI.construct_prefix_v6(prefix_str))


if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Link State attribute"""

import unittest
from yabgp.message.attribute.linkstate.linkstate import LinkState


class TestLinkDelay(unittest.TestCase):

    def test_unpack(self):
        self.maxDiff = None
        data_bin = b"\x04\x04\x00\x04\x03\x03\x03\x03\x04\x40\x00\x04\x00\x00\x00\x00" \
                   b"\x04\x41\x00\x04\x4c\xee\x6b\x28\x04\x42\x00\x04\x00\x00\x00\x00" \
                   b"\x04\x43\x00\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" \
                   b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" \
                   b"\x00\x00\x00\x00\x04\x44\x00\x04\x00\x00\x00\x01\x04\x47\x00\x03" \
                   b"\x00\x00\x01\x04\x4b\x00\x07\x60\x00\x00\x00\x00\x5d\xc1\x01\x0b" \
                   b"\x00\x02\x01\x0a\x04\x5a\x00\x04\x00\x0f\x42\x40\x04\x5b\x00\x08" \
                   b"\x00\x0f\x42\x40\x00\x0f\x42\x40\x04\x5c\x00\x04\x00\x00\x00\x00" \
                   b"\x04\x5d\x00\x04\x11\x11\x11\x11\x04\x5e\x00\x04\x11\x11\x11\x11" \
                   b"\x04\x5f\x00\x04\x11\x11\x11\x11\x04\x60\x00\x04\x11\x11\x11\x11"

        data_dict = {29: [{
            "type": "local_router_id",
            "value": "3.3.3.3"
        }, {
            "type": "admin_group",
            "value": 0
        }, {
            "type": "max_bandwidth",
            "value": 125000000.0
        }, {
            "type": "max_rsv_bandwidth",
            "value": 0.0
        }, {
            "type": "unrsv_bandwidth",
            "value": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        }, {
            "type": "te_metric",
            "value": 1
        }, {
            "type": "igp_metric",
            "value": 1
        }, {
            "type": "adj_sid",
            "value": {
                "flags": {
                    "P": 0,
                    "B": 0,
                    "L": 1,
                    "G": 0,
                    "V": 1
                },
                "value": 24001,
                "weight": 0
            }
        }, {
            "type": "link_msd",
            "value": {
                "type": 1,
                "value": 10
            }
        }, {
            "type": "unidirect_link_delay",
            "value": 1000000
        }, {
            "type": "min_max_unidirect_link_delay",
            "value": {
                "max_link_delay": 1000000,
                "min_link_delay": 1000000
            }
        }, {
            "type": "unidirect_delay_var",
            "value": 0
        }, {
            "type": "unidirect_packet_loss",
            "value": 286331153
        }, {
            "type": "unidirect_residual_bw",
            "value": 286331153
        }, {
            "type": "unidirect_avail_bw",
            "value": 286331153
        }, {
            "type": "unidirect_bw_util",
            "value": 286331153
        }]}

        self.assertEqual(data_dict, LinkState.unpack(data_bin).dict())



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/rfc7752#section-3.3.2.2
#  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |              Type             |             Length            |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |L|R|  Reserved |
# +-+-+-+-+-+-+-+-+

# +------------+------------------------------------------+-----------+
# |    Bit     | Description                              | Reference |
# +------------+------------------------------------------+-----------+
# |    'L'     | Label Distribution Protocol (LDP)        | [RFC5036] |
# |    'R'     | Extension to RSVP for LSP Tunnels        | [RFC3209] |
# |            | (RSVP-TE)                                |           |
# | 'Reserved' | Reserved for future use                  |           |
# +------------+------------------------------------------+-----------+


@LinkState.register()
class MplsMask(TLV):

    TYPE = 1094  # https://tools.ietf.org/html/rfc7752#section-3.3.2.2
    TYPE_STR = "mpls_mask"

    @classmethod
    def unpack(cls, value):
        """
        """
        value = ord(value[0:1])
        L = value >> 7
        R = (value << 1) % 256 >> 7
        return cls(value={'L': L, 'R': R})



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test MPLS VPN """

import unittest

from yabgp.message.attribute.nlri.mpls_vpn import MPLSVPN


class TestMPLSVPN(unittest.TestCase):

    def test_parse_mpls_label_stack(self):
        label_bin = b'\x00\x01\x41'
        self.assertEqual([20], MPLSVPN.parse_mpls_label_stack(label_bin))
        label_bin = b'\x00\x01\x10\x00\x01\x31'
        self.assertEqual([17, 19], MPLSVPN.parse_mpls_label_stack(label_bin))
        label_bin = b'\x80\x00\x00'
        self.assertEqual([524288], MPLSVPN.parse_mpls_label_stack(label_bin))

    def test_construct_mpls_label_stack(self):
        label_bin = b'\x00\x01\x41'
        self.assertEqual(label_bin, MPLSVPN.construct_mpls_label_stack(labels=[20]))

    def test_parse_and_construct_rd_type0(self):
        # for rd type 0
        rd_bin = b'\x00\x00\xfd\xea\x00\x00\x00\x01'
        rd_parsed = '65002:1'
        self.assertEqual(rd_parsed, MPLSVPN.parse_rd(rd_bin))
        self.assertEqual(rd_bin, MPLSVPN.construct_rd(rd_parsed))

    def test_parse_and_construct_rd_type1(self):

        rd_bin = b'\x00\x01\xac\x11\x00\x03\x00\x02'
        rd_parsed = '172.17.0.3:2'
        self.assertEqual(rd_parsed, MPLSVPN.parse_rd(rd_bin))
        self.assertEqual(rd_bin, MPLSVPN.construct_rd(rd_parsed))

    def test_parse_and_construct_rd_type2(self):
        rd_bin = b'\x00\x02\x00\x01\x00\x00\x00\x02'
        rd_parsed = '65536:2'
        self.assertEqual(rd_parsed, MPLSVPN.parse_rd(rd_bin))
        self.assertEqual(rd_bin, MPLSVPN.construct_rd(rd_parsed))

if __name__ == '__main__':
    unittest.main()



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import binascii

from yabgp.tlv import TLV
from ..linkstate import LinkState


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |               Type            |              Length           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# | Flags         |     Weight    |             Reserved          |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                   SID/Label/Index (variable)                  |
# +---------------------------------------------------------------+

# isis
# 0 1 2 3 4 5 6 7
# +-+-+-+-+-+-+-+-+
# |F|B|V|L|S|P|   |
# +-+-+-+-+-+-+-+-+

# ospf
# 0 1 2 3 4 5 6 7
# +-+-+-+-+-+-+-+-+
# |B|V|L|G|P|     |
# +-+-+-+-+-+-+-+-+

# ospf-v3
# 0 1 2 3 4 5 6 7
# +-+-+-+-+-+-+-+-+
# |B|V|L|G|P|     |
# +-+-+-+-+-+-+-+-+


@LinkState.register()
class AdjSegID(TLV):
    """
    Adjacency Segment id
    """
    TYPE = 1099
    TYPE_STR = 'adj_sid'

    @classmethod
    def unpack(cls, data, pro_id):
        flags = ord(data[0:1])
        flag = {}
        if pro_id in [1, 2]:
            flag['F'] = flags >> 7
            flag['B'] = (flags << 1) % 256 >> 7
            flag['V'] = (flags << 2) % 256 >> 7
            flag['L'] = (flags << 3) % 256 >> 7
            flag['S'] = (flags << 4) % 256 >> 7
            flag['P'] = (flags << 5) % 256 >> 7
        else:  # 3, 6
            flag['B'] = flags >> 7
            flag['V'] = (flags << 1) % 256 >> 7
            flag['L'] = (flags << 2) % 256 >> 7
            flag['G'] = (flags << 3) % 256 >> 7
            flag['P'] = (flags << 4) % 256 >> 7
        weight = ord(data[1:2])
        return cls(value={"flags": flag, "weight": weight, "value": int(binascii.b2a_hex(data[4:]), 16)})
        # return cls(value=int(binascii.b2a_hex(data[4:]), 16))



# Copyright 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from .linkstate import LinkState  # noqa
from .node.local_router_id import LocalRouterID   # noqa
from .node.name import NodeName   # noqa
from .node.isisarea import ISISArea   # noqa
from .node.sr_capabilities import SRCapabilities  # noqa
from .node.sr_algorithm import SRAlgorithm  # noqa
from .node.node_msd import NodeMSD  # noqa
from .node.nodeflags import NodeFlags  # noqa
from .node.opa_node_attr import OpaNodeAttr  # noqa
from .node.sid_or_label import SIDorLabel   # noqa
from .node.srlb import SRLB   # noqa
from .link.admingroup import AdminGroup   # noqa
from .link.remote_router_id import RemoteRouterID   # noqa
from .link.max_bw import MaxBandwidth   # noqa
from .link.max_rsv_bw import MaxResvBandwidth   # noqa
from .link.unsrv_bw import UnrsvBandwidth    # noqa
from .link.te_metric import TeMetric   # noqa
from .link.link_name import LinkName   # noqa
from .link.igp_metric import IGPMetric   # noqa
from .link.adj_seg_id import AdjSegID   # noqa
from .link.link_identifiers import LinkIdentifiers  # noqa
from .link.link_msd import LinkMSD  # noqa
from .link.lan_adj_sid import LanAdjSegID  # noqa
from .link.srlg import SRLGList  # noqa
from .link.mplsmask import MplsMask  # noqa
from .link.protection_type import ProtectionType  # noqa
from .link.opa_link_attr import OpaLinkAttr  # noqa
from .link.peer_node_sid import PeerNodeSID   # noqa
from .link.peer_adj_sid import PeerAdjSID    # noqa
from .link.peer_set_sid import PeerSetSID    # noqa
from .link.unidirect_link_delay import UnidirectLinkDelay    # noqa
from .link.min_max_link_delay import MinMaxUnidirectLinkDelay    # noqa
from .link.unidirect_delay_var import UnidirectDelayVar    # noqa
from .link.unidirect_packet_loss import UnidirectPacketLoss    # noqa
from .link.unidirect_residual_bw import UnidirectResidualBw    # noqa
from .link.unidirect_avail_bw import UnidirectAvailBw    # noqa
from .link.unidirect_bw_util import UnidirectBwUtil    # noqa
from .prefix.prefix_metric import PrefixMetric   # noqa
from .prefix.prefix_sid import PrefixSID  # noqa
from .prefix.prefix_igp_attr import PrefixIGPAttr  # noqa
from .prefix.src_router_id import SrcRouterID  # noqa
from .prefix.igpflags import IGPFlags  # noqa
from .prefix.igp_route_tag_list import IGPRouteTagList  # noqa
from .prefix.ext_igp_route_tag_list import ExtIGPRouteTagList  # noqa
from .prefix.ospf_forward_addr import OspfForwardingAddr  # noqa



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.tlv import TLV
from ..linkstate import LinkState

#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                  Link Local Identifier                        |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                  Link Remote Identifier                       |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


@LinkState.register()
class LinkIdentifiers(TLV):
    """
    link local/remote identifiers
    """
    TYPE = 258  # RFC 7752
    TYPE_STR = 'link_identifiers'

    @classmethod
    def unpack(cls, data):
        local_identifier = struct.unpack('!I', data[:4])[0]
        remote_identifier = struct.unpack('!I', data[4:])[0]

        return cls(value={'local_identifier': local_identifier, 'remote_identifier': remote_identifier})



# Copyright (c) 2007 by Mark Bergsma <mark@nedworks.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Twisted Factory, BGP implementation.
"""

import logging
import socket
import platform
import struct
import sys

import netaddr
from twisted.internet import protocol
from twisted.internet import reactor
from oslo_config import cfg

from yabgp.core.protocol import BGP
from yabgp.core.fsm import FSM
from yabgp.common import constants as bgp_cons
from yabgp.common.afn import AFNUM_INET

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class BGPFactory(protocol.Factory):
    """Base factory for creating BGP protocol instances."""

    protocol = BGP
    FSM = FSM

    def buildProtocol(self, addr):
        """Builds a BGPProtocol instance.

        :param addr : address used for buliding protocol.
        """
        return protocol.Factory.buildProtocol(self, addr)

    def startedConnecting(self, connector):
        """Called when a connection attempt has been initiated.

        :param connector : Twisted connector
        """
        pass

    def clientConnectionLost(self, connector, reason):
        """ Called when a TCP client connection was lost.

        :param connector : Twisted connector
        :param reason : connection faied reason.
        """
        LOG.info("Client connection lost:%s", reason.getErrorMessage())


class BGPPeering(BGPFactory):
    """Class managing a BGP session with a peer.
    One connection, One BGPPeering class.
    """

    def __init__(self, myasn=None, myaddr=None, peerasn=None, peeraddr=None,
                 afisafi=None, md5=None, handler=None):
        """Initial a BGPPeering instance.

        :param myasn: local bgp as number.
        :param myaddr: local ip address.
        :param peerasn: remote bgp peer as number
        :param peeraddr: remote peer ip address.
        :param msgpath: the path to store bgp message file.
        :param afisafi: afi and safi
        :param md5: TCP md5 string
        """
        LOG.info('Init BGPPeering for peer %s', peeraddr)
        self.my_asn = myasn
        self.my_addr = myaddr
        self.peer_addr = peeraddr
        self.peer_id = None
        self.bgp_id = None
        self.peer_asn = peerasn
        self.afi_safi = afisafi
        self.md5 = md5

        self.status = False
        self.fsm = BGPFactory.FSM(self)
        self.handler = handler

        # reference to the BGPProtocol instance in ESTAB state
        self.estab_protocol = None

    def buildProtocol(self, addr):

        """Builds a BGP protocol instance

        :param addr: IP address used for building protocol.
        """
        LOG.info("[%s]Building a new BGP protocol instance", self.peer_addr)

        p = BGPFactory.buildProtocol(self, addr)
        if p is not None:
            self._initProtocol(p, addr)
        self.estab_protocol = p
        return p

    def _initProtocol(self, protocol, addr):

        """Initializes a BGPProtocol instance

        :param protocol: twisted Protocol
        :param addr: ip address
        """

        protocol.bgp_peering = self

        # Hand over the FSM
        protocol.fsm = self.fsm
        protocol.fsm.protocol = protocol

        if addr.port == bgp_cons.PORT:
            protocol.fsm.state = bgp_cons.ST_CONNECT
        else:
            protocol.fsm.state = bgp_cons.ST_ACTIVE

    def clientConnectionFailed(self, connector, reason):

        """Called when the outgoing connection failed.

        :param connector: Twisted connector
        :param reason: connection failed reason
        """

        error_msg = "[%s]Client connection failed: %s" % (self.peer_addr, reason.getErrorMessage())
        self.handler.on_connection_failed(self.peer_addr, reason.getErrorMessage())
        LOG.info(error_msg)
        # There is no protocol instance yet at this point.
        # Catch a possible NotificationException
        try:
            self.fsm.connection_failed()
        except Exception as e:
            LOG.info("[%s]Client connection failed: %s", self.peer_addr, e)

    def automatic_start(self, idle_hold=False):

        """BGP AutomaticStart event (event 3)

        :param idle_hold: flag represents used Idle Hold
        """
        if self.fsm.state == bgp_cons.ST_IDLE:
            if self.fsm.automatic_start(idle_hold):
                self.status = True
                # Create outbound connection as a client
                self.connect()

    def manual_start(self, idle_hold=False):
        if self.fsm.state == bgp_cons.ST_IDLE:
            if self.fsm.manual_start(idle_hold=idle_hold):
                self.status = True
                self.connect()

    def manual_stop(self):

        """BGP ManualStop event (event 2) Returns a DeferredList that
        will fire once the connection(s) have closed"""

        try:
            self.estab_protocol.fsm.manual_stop()
        except Exception as e:
            LOG.error(e)

    def connection_closed(self, pro, disconnect=False):
        """
        Called by FSM or Protocol when the BGP connection has been closed.

        :param pro: twisted protocol
        :param disconnect: the status of connection
        """

        LOG.info("[%s]Connection closed", self.peer_addr)
        if pro is not None:
            # Connection succeeded previously, protocol exists
            # Remove the protocol, if it exists
            if pro is self.estab_protocol:
                self.estab_protocol = None
                # self.fsm should still be valid and set to ST_IDLE
                self.fsm.state = bgp_cons.ST_IDLE

        if self.fsm.allow_automatic_start:
            self.automatic_start(idle_hold=True)

    def connect_retry(self):

        """Called by FSM when we should reattempt to connect.
        """
        try:
            self.connect()
        except Exception as e:
            LOG.error(e)
            import traceback
            LOG.debug(traceback.format_exc())

    def set_peer_id(self, bgp_id):
        """
        Should be called when an Open message was received from a peer.
        Sets the BGP identifier of the peer if it wasn't set yet. If the
        new peer id is unequal to the existing one, CEASE all connections.
        s
        :param bgp_id: BGP ID
        """

        if self.peer_id is None:
            self.peer_id = bgp_id
            LOG.info('Set BGP peer id %s', bgp_id)
        elif self.peer_id != bgp_id:
            # Ouch, schizophrenia. The BGP id of the peer is unequal to
            # the ids of current and/or previous sessions. Close all
            # connections.
            self.peer_id = None

    def connect(self):

        """Initiates a TCP client connection to the peer. Should only be called from
        BGPPeering or FSM, otherwise use manualStart() instead.
        """
        # DEBUG
        LOG.info("(Re)connect to %s", self.peer_addr)

        if self.fsm.state != bgp_cons.ST_ESTABLISHED:

            connector = reactor.connectTCP(
                host=self.peer_addr,
                port=bgp_cons.PORT,
                factory=self,
                timeout=30,
                bindAddress=(self.my_addr, 0))
            if isinstance(self.md5, str) and self.md5:
                md5sig = self.get_tcp_md5sig(self.md5, self.peer_addr, bgp_cons.PORT)
                if md5sig:
                    connector.transport.getHandle().setsockopt(
                        socket.IPPROTO_TCP, bgp_cons.TCP_MD5SIG, md5sig)
                else:
                    sys.exit()
            return True
        else:
            return False

    @staticmethod
    def get_tcp_md5sig(md5_str, host, port):
        """set tcp md5
        """
        os_type = platform.system()
        if os_type != 'Linux':
            LOG.error('YABGP has no MD5 support for %s', os_type)
            return None

        # address family
        if netaddr.IPAddress(host).version == 4:
            # ipv4 address
            afi = AFNUM_INET
        elif netaddr.IPAddress(host).version == 6:
            # ipv6 address
            LOG.error("Does not support ipv6 address family")
            sys.exit()
        else:
            afi = None
            LOG.error('Peer address is not a valid ipv4/ipv6 address')
        try:
            n_port = socket.htons(port)
            if afi == AFNUM_INET:
                n_addr = socket.inet_pton(socket.AF_INET, host)
                tcp_md5sig = 'HH4s%dx2xH4x%ds' % (
                    bgp_cons.SS_PADSIZE_IPV4, bgp_cons.TCP_MD5SIG_MAXKEYLEN)
                md5sig = struct.pack(
                    tcp_md5sig, socket.AF_INET, n_port, n_addr, len(md5_str), md5_str.encode())
                return md5sig
            else:
                return None
        except socket.error as e:
            LOG.error('This linux machine does not support TCP_MD5SIG: (%s)', str(e))
            return None



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
from yabgp.message.attribute.linkstate.linkstate import LinkState
from yabgp.tlv import TLV

# https://tools.ietf.org/html/rfc7752#section-3.3.3.2
#  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |              Type             |             Length            |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# //                    Route Tags (one or more)                 //
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


@LinkState.register()
class IGPRouteTagList(TLV):

    TYPE = 1153  # https://tools.ietf.org/html/rfc7752#section-3.3.3.2
    TYPE_STR = "igp_route_tag_list"

    @classmethod
    def unpack(cls, value):
        """
        """
        results = []
        while value:
            results.append(struct.unpack('!I', value[:4])[0])
            value = value[4:]
        return cls(value=results)



# Copyright 2015-2018 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.tlv import TLV
from ..linkstate import LinkState


@LinkState.register(_type=1050)
@LinkState.register(_type=266)
class NodeMSD(TLV):
    """
    node msd
    """
    # https://tools.ietf.org/html/draft-tantsura-idr-bgp-ls-segment-routing-msd-05#section-3
    TYPE_STR = 'node_msd'

    @classmethod
    def unpack(cls, data):
        _type, value = struct.unpack('!BB', data)
        return cls(value={"type": _type, "value": value})



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
import binascii

import netaddr
from yabgp.common import afn
from yabgp.common import safn
from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri import NLRI
from yabgp.message.attribute.nlri.mpls_vpn import MPLSVPN


class EVPN(NLRI):
    """
      The format of the EVPN NLRI is as follows:
    +-----------------------------------+
    |    Route Type (1 octet)           |
    +-----------------------------------+
    |     Length (1 octet)              |
    +-----------------------------------+
    | Route Type specific (variable)    |
    +-----------------------------------+
    """

    @classmethod
    def parse(cls, nlri_data):
        nlri_list = []
        while nlri_data:
            route_type = ord(nlri_data[0:1])
            offset = ord(nlri_data[1:2])
            route_value = nlri_data[2: offset + 2]
            route = {}
            if route_type == bgp_cons.BGPNLRI_EVPN_ETHERNET_AUTO_DISCOVERY:
                route = EthernetAutoDiscovery.parse(route_value)
            elif route_type == bgp_cons.BGPNLRI_EVPN_MAC_IP_ADVERTISEMENT:
                route = MacIPAdvertisment.parse(route_value)
            elif route_type == bgp_cons.BGPNLRI_EVPN_INCLUSIVE_MULTICAST_ETHERNET_TAG:
                route = InclusiveMulticastEthernetTag.parse(route_value)
            elif route_type == bgp_cons.BGPNLRI_EVPN_ETHERNET_SEGMENT:
                route = EthernetSegment.parse(route_value)
            elif route_type == bgp_cons.BGPNLRI_EVPN_IP_ROUTE_PREFIX:
                route = IPRoutePrefix.parse(route_value)
            if route:
                nlri_list.append({
                    'type': route_type,
                    'value': route
                })
            nlri_data = nlri_data[offset + 2:]
        return nlri_list

    @classmethod
    def construct(cls, nlri_list):
        nlri_list_hex = b''
        for nlri in nlri_list:
            nlri_hex = b''
            if nlri['type'] == bgp_cons.BGPNLRI_EVPN_ETHERNET_AUTO_DISCOVERY:
                nlri_hex = EthernetAutoDiscovery.construct(value=nlri['value'])
            elif nlri['type'] == bgp_cons.BGPNLRI_EVPN_MAC_IP_ADVERTISEMENT:
                nlri_hex = MacIPAdvertisment.construct(value=nlri['value'])
            elif nlri['type'] == bgp_cons.BGPNLRI_EVPN_INCLUSIVE_MULTICAST_ETHERNET_TAG:
                nlri_hex = InclusiveMulticastEthernetTag.construct(value=nlri['value'])
            elif nlri['type'] == bgp_cons.BGPNLRI_EVPN_ETHERNET_SEGMENT:
                nlri_hex = EthernetSegment.construct(value=nlri['value'])
            elif nlri['type'] == bgp_cons.BGPNLRI_EVPN_IP_ROUTE_PREFIX:
                nlri_hex = IPRoutePrefix.construct(value=nlri['value'])
            if nlri_hex:
                nlri_list_hex += struct.pack('!2B', nlri['type'], len(nlri_hex)) + nlri_hex
        return nlri_list_hex

    @staticmethod
    def signal_evpn_overlay(attr_dict):
        """
        draft-ietf-bess-evpn-overlay-10 changes label encoding if EVPN and encapsulation EC set

        :param attr_dict: bgp attribute dictionary
        """
        evpn_overlay = {'evpn': False, 'encap_ec': False}
        try:
            afi_safi = tuple(attr_dict.get(bgp_cons.BGPTYPE_MP_REACH_NLRI).get('afi_safi'))
            community_ext = attr_dict.get(bgp_cons.BGPTYPE_EXTENDED_COMMUNITY)
        except:
            return evpn_overlay
        if afi_safi == (afn.AFNUM_L2VPN, safn.SAFNUM_EVPN):
            evpn_overlay['evpn'] = True
        if community_ext:
            for ec in community_ext:
                if bgp_cons.BGP_EXT_COM_DICT['encapsulation'] == ec[0]:
                    evpn_overlay['encap_ec'] = True
                    evpn_overlay['encap_value'] = int(ec[1])
        return evpn_overlay


class EthernetAutoDiscovery(MPLSVPN):
    """
    +---------------------------------------+
    |  Route Distinguisher (RD) (8 octets)  |
    +---------------------------------------+
    |Ethernet Segment Identifier (10 octets)|
    +---------------------------------------+
    |  Ethernet Tag ID (4 octets)           |
    +---------------------------------------+
    |  MPLS Label (3 octets)                |
    +---------------------------------------+

    """
    @classmethod
    def parse(cls, value, iswithdraw=False):
        route = dict()
        route['rd'] = cls.parse_rd(value[0:8])
        offset = 8
        route['esi'] = int(binascii.b2a_hex(value[offset: offset+10]), 16)
        offset += 10
        # ethernet tag id
        route['eth_tag_id'] = struct.unpack('!I', value[offset: offset+4])[0]
        offset += 4
        route['label'] = MPLSVPN.parse_mpls_label_stack(value[offset:])
        return route

    @classmethod
    def construct(cls, value, iswithdraw=False):
        # rd
        value_hex = b''
        value_hex += MPLSVPN.construct_rd(value['rd'])
        # esi
        value_hex += b'\x00\x00' + struct.pack('!d', value['esi'])
        # ethernet tag
        value_hex += struct.pack('!I', value['eth_tag_id'])
        value_hex += MPLSVPN.construct_mpls_label_stack(value['label'])
        return value_hex


class MacIPAdvertisment(MPLSVPN):
    """
    +---------------------------------------+
    |  RD (8 octets)                        |
    +---------------------------------------+
    |Ethernet Segment Identifier (10 octets)|
    +---------------------------------------+
    |  Ethernet Tag ID (4 octets)           |
    +---------------------------------------+
    |  MAC Address Length (1 octet)         |
    +---------------------------------------+
    |  MAC Address (6 octets)               |
    +---------------------------------------+
    |  IP Address Length (1 octet)          |
    +---------------------------------------+
    |  IP Address (0, 4, or 16 octets)      |
    +---------------------------------------+
    |  MPLS Label1 (3 octets)               |
    +---------------------------------------+
    |  MPLS Label2 (0 or 3 octets)          |
    +---------------------------------------+
    """

    @classmethod
    def parse(cls, value, iswithdraw=False):
        route = dict()
        # rd
        offset = 8
        route['rd'] = cls.parse_rd(value[0:offset])
        # esi
        route['esi'] = int(binascii.b2a_hex(value[offset: offset+10]), 16)
        offset += 10
        # ethernet tag id
        route['eth_tag_id'] = struct.unpack('!I', value[offset: offset+4])[0]
        offset += 5
        # mac address
        route['mac'] = str(netaddr.EUI(int(binascii.b2a_hex(value[offset: offset+6]), 16)))
        offset += 6
        ip_addr_len = ord(value[offset: offset + 1])
        offset += 1
        # ip address
        if ip_addr_len != 0:
            route['ip'] = str(netaddr.IPAddress(
                int(binascii.b2a_hex(value[offset: offset + int(ip_addr_len / 8)]), 16)))
            offset += int(ip_addr_len / 8)
        # label
        route['label'] = MPLSVPN.parse_mpls_label_stack(value[offset:])
        return route

    @classmethod
    def construct(cls, value, iswithdraw=False):
        # rd
        value_hex = b''
        value_hex += MPLSVPN.construct_rd(value['rd'])
        # esi
        value_hex += b'\x00\x00' + struct.pack('!d', value['esi'])
        # ethernet tag
        value_hex += struct.pack('!I', value['eth_tag_id'])
        # mac address len and address
        mac_hex = b''.join([struct.pack('!B', (int(i, 16))) for i in value['mac'].split("-")])
        value_hex += struct.pack('!B', len(mac_hex) * 8) + mac_hex
        # ip address len and address
        if value.get('ip'):
            ip_hex = netaddr.IPAddress(value['ip']).packed
            value_hex += struct.pack('!B', len(ip_hex) * 8) + ip_hex
        else:
            value_hex += b'\x00'
        if value.get('label'):
            value_hex += MPLSVPN.construct_mpls_label_stack(value['label'])
        return value_hex


class InclusiveMulticastEthernetTag(MPLSVPN):
    """
   +---------------------------------------+
   |  RD (8 octets)                        |
   +---------------------------------------+
   |  Ethernet Tag ID (4 octets)           |
   +---------------------------------------+
   |  IP Address Length (1 octet)          |
   +---------------------------------------+
   |  Originating Router's IP Address      |
   |          (4 or 16 octets)             |
   +---------------------------------------+
    """

    @classmethod
    def parse(cls, value, iswithdraw=False):
        route = dict()
        offset = 8
        route['rd'] = MPLSVPN.parse_rd(value[0:offset])
        route['eth_tag_id'] = struct.unpack('!I', value[offset: offset+4])[0]
        offset += 4
        ip_addr_len = ord(value[offset: offset + 1])
        offset += 1
        # ip address
        if ip_addr_len != 0:
            route['ip'] = str(netaddr.IPAddress(int(binascii.b2a_hex(value[offset: int(offset+ip_addr_len / 8)]), 16)))
        return route

    @classmethod
    def construct(cls, value, iswithdraw=False):
        # rd
        value_hex = b''
        value_hex += MPLSVPN.construct_rd(value['rd'])
        value_hex += struct.pack('!I', value['eth_tag_id'])
        # ip address len and address
        if value.get('ip'):
            ip_hex = netaddr.IPAddress(value['ip']).packed
            value_hex += struct.pack('!B', len(ip_hex) * 8) + ip_hex
        else:
            value_hex += b'\x00'
        return value_hex


class EthernetSegment(MPLSVPN):
    """
   +---------------------------------------+
   |  RD (8 octets)                        |
   +---------------------------------------+
   |Ethernet Segment Identifier (10 octets)|
   +---------------------------------------+
   |  IP Address Length (1 octet)          |
   +---------------------------------------+
   |  Originating Router's IP Address      |
   |          (4 or 16 octets)             |
   +---------------------------------------+

    """
    @classmethod
    def parse(cls, value, iswithdraw=False):
        route = dict()
        offset = 8
        route['rd'] = MPLSVPN.parse_rd(value[0:offset])
        # esi
        route['esi'] = int(binascii.b2a_hex(value[offset: offset+10]), 16)
        offset += 10
        ip_addr_len = ord(value[offset: offset + 1])
        offset += 1
        # ip address
        if ip_addr_len != 0:
            route['ip'] = str(netaddr.IPAddress(int(binascii.b2a_hex(value[offset: offset+ip_addr_len / 8]), 16)))
        return route

    @classmethod
    def construct(cls, value, iswithdraw=False):
        # rd
        value_hex = b''
        value_hex += MPLSVPN.construct_rd(value['rd'])
        # esi
        value_hex += b'\x00\x00' + struct.pack('!d', value['esi'])
        # ip address len and address
        if value.get('ip'):
            ip_hex = netaddr.IPAddress(value['ip']).packed
            value_hex += struct.pack('!B', len(ip_hex) * 8) + ip_hex
        else:
            value_hex += b'\x00'
        return value_hex


class IPRoutePrefix(MPLSVPN):
    """
    # http://tools.ietf.org/html/draft-ietf-bess-evpn-prefix-advertisement-01
    +---------------------------------------+
    |      RD   (8 octets)                  |
    +---------------------------------------+
    |Ethernet Segment Identifier (10 octets)|
    +---------------------------------------+
    |  Ethernet Tag ID (4 octets)           |
    +---------------------------------------+
    |  IP Prefix Length (1 octet)           |
    +---------------------------------------+
    |  IP Prefix (4 or 16 octets)           |
    +---------------------------------------+
    |  GW IP Address (4 or 16 octets)       |
    +---------------------------------------+
    |  MPLS Label (3 octets)                |
    +---------------------------------------+
    """
    @classmethod
    def parse(cls, value, iswithdraw=False):
        route = dict()
        offset = 8
        route['rd'] = cls.parse_rd(value[0:offset])
        # esi
        route['esi'] = int(binascii.b2a_hex(value[offset: offset+10]), 16)
        offset += 10

        route['eth_tag_id'] = struct.unpack('!I', value[offset: offset+4])[0]
        offset += 4

        ip_addr_len = ord(value[offset: offset + 1])
        offset += 1

        value = value[offset:]
        # The IP Prefix Length can be set to a value between 0 and 32
        #   (bits) for ipv4 and between 0 and 128 for ipv6.
        # The IP Prefix will be a 32 or 128-bit field (ipv4 or ipv6).

        # # ip address
        if len(value) == 11:
            # ipv4
            offset = 4
        elif len(value) == 35:
            # ipv6
            offset = 16

        route['prefix'] = '%s/%s' % (str(netaddr.IPAddress(int(binascii.b2a_hex(value[0: offset]), 16))), ip_addr_len)
        value = value[offset:]
        route['gateway'] = str(netaddr.IPAddress(int(binascii.b2a_hex(value[0: offset]), 16)))
        value = value[offset:]

        route['label'] = cls.parse_mpls_label_stack(value)
        return route

    @classmethod
    def construct(cls, value, iswithdraw=False):
        value_hex = b''
        value_hex += cls.construct_rd(value['rd'])
        value_hex += b'\x00\x00' + struct.pack('!d', value['esi'])
        value_hex += struct.pack('!I', value['eth_tag_id'])
        value_hex += struct.pack('!B', int(value['prefix'].split('/')[1]))
        value_hex += netaddr.IPAddress(value['prefix'].split('/')[0]).packed
        value_hex += netaddr.IPAddress(value['gateway']).packed
        value_hex += cls.construct_mpls_label_stack(value['label'])
        return value_hex
